/*******************************************************************************************/
/**                                                                                       **/
/** COPYRIGHT (C) 2012, Systemyde International Corporation, ALL RIGHTS RESERVED          **/
/**                                                                                       **/
/** 41CL level shifters                                               Rev 0.0  12/21/2012 **/
/**                                                                                       **/
/*******************************************************************************************/
module board (data_bus, isa_bus, fi_bus, fo_bus, ph1, ph2, pwo, sync, DATA_BUS_EN,
              DATA_BUS_HI, FO_BUS_HI, ISA_BUS_EN, ISA_BUS_HI, PH1_HI, PH2_HI, PRECHRG,
              PWO_HI, SYNC_DRV);

  inout         data_bus;      /* memory/peripheral data bus                          bus  */
  inout         isa_bus;       /* instruction address/data bus                        bus  */
  inout         fi_bus;        /* flag input bus                                      bus  */
  input         DATA_BUS_EN;   /* memory/peripheral data bus enable                   fpga */
  input         DATA_BUS_HI;   /* memory/peripheral data bus hi drive                 fpga */
  input         FO_BUS_HI;     /* flag output bus hi drive                            fpga */
  input         ISA_BUS_EN;    /* instruction address/data bus enable                 fpga */
  input         ISA_BUS_HI;    /* instruction address/data bus hi drive               fpga */
  input         PH1_HI;        /* clock phase 1 hi drive                              fpga */
  input         PH2_HI;        /* clock phase 2 hi drive                              fpga */
  input         PRECHRG;       /* precharge                                           fpga */
  input         PWO_HI;        /* power off                                           fpga */
  input         SYNC_DRV;      /* sync drive                                          cpld */
  output        fo_bus;        /* flag output bus                                     bus  */
  output        ph1;           /* clock phase 1                                       bus  */
  output        ph2;           /* clock phase 2                                       bus  */
  output        pwo;           /* power off                                           bus  */
  output        sync;          /* synchronization output                              bus  */

  /*****************************************************************************************/
  /*                                                                                       */
  /* signal declarations                                                                   */
  /*                                                                                       */
  /*****************************************************************************************/
  wire        fo_bus, ph1, ph2, pwo, sync;
  reg         data_bus_reg, isa_bus_reg;

  /*****************************************************************************************/
  /*                                                                                       */
  /* 6V signals                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (ISA_BUS_EN or ISA_BUS_HI) begin
    case ({ISA_BUS_EN, ISA_BUS_HI})
      2'b10:   isa_bus_reg = 1'b0;
      2'b11:   isa_bus_reg = 1'b1;
      default: isa_bus_reg = 1'bz;
      endcase
    end

  always @ (DATA_BUS_EN or DATA_BUS_HI) begin
    case ({DATA_BUS_EN, DATA_BUS_HI})
      2'b10:   data_bus_reg = 1'b0;
      2'b11:   data_bus_reg = 1'b1;
      default: data_bus_reg = 1'bz;
      endcase
    end

  assign data_bus  = data_bus_reg;                         /* data bus                     */
  assign fi_bus    = (PRECHRG) ? 1'b1 : 1'bz;              /* flag input                   */
  assign fo_bus    = FO_BUS_HI;                            /* flag output                  */
  assign isa_bus   = isa_bus_reg;                          /* inst/addr bus                */
  assign ph1       = PH1_HI;                               /* phase 1                      */
  assign ph2       = PH2_HI;                               /* phase 2                      */
  assign pwo       = PWO_HI;                               /* power off                    */
  assign sync      = SYNC_DRV;                             /* synchronization              */

  pulldown(data_bus);
  pulldown(isa_bus);
  pullup(fi_bus);

  endmodule






