/*******************************************************************************************/
/**                                                                                       **/
/** COPYRIGHT (C) 2011, Systemyde International Corporation, ALL RIGHTS RESERVED          **/
/**                                                                                       **/
/** nut cpu                                                           Rev 0.0  03/14/2011 **/
/**                                                                                       **/
/*******************************************************************************************/
module nut_cpu (chkkb_inst, cycle_reg, data_drv, data_out, dout_reg, fo_bus, isa_drv,
                isa_nxt, isa_out, jmpldi_reg, ld_cycle, ld_inst, powoff_inst, rstkb_inst,
                selprf_reg, state_reg, sync_raw, sync_reg, te_a14, data_in, dpwo, fi_bus,
                isa_in, ky_flag, ky_data, lld, ph2, rstb);

  input         data_in;       /* memory/peripheral data input                             */
  input         isa_in;        /* instruction address/data input                           */
  input         dpwo;          /* display power on request                                 */
  input         fi_bus;        /* flag input bus                                           */
  input         ky_flag;       /* keyboard flag                                            */
  input         ky_data;       /* keyboard data                                            */
  input         lld;           /* low-level detect                                         */
  input         ph2;           /* processor clock                                          */
  input         rstb;          /* master reset                                             */
  output        chkkb_inst;    /* chk kb instruction                                       */
  output        data_drv;      /* memory/peripheral data direction                         */
  output        data_out;      /* memory/peripheral data output                            */
  output        fo_bus;        /* flag output bus                                          */
  output        isa_drv;       /* isa drive enable                                         */
  output        isa_nxt;       /* next isa output data                                     */
  output        isa_out;       /* isa output data                                          */
  output        jmpldi_reg;    /* jmp/jsb/ldi/cxisa latched                                */
  output        ld_cycle;      /* load machine cycle                                       */
  output        ld_inst;       /* load instruction                                         */
  output        powoff_inst;   /* power off instruction                                    */
  output        rstkb_inst;    /* rst kb instruction                                       */
  output        selprf_reg;    /* selprf latched                                           */
  output        sync_raw;      /* raw sync signal                                          */
  output        sync_reg;      /* sync signal                                              */
  output        te_a14;        /* time enable: addr 14                                     */
  output  [3:0] cycle_reg;     /* clock cycle                                              */
  output  [3:0] state_reg;     /* machine state                                            */
  output [55:0] dout_reg;      /* stable data register                                     */

  /*****************************************************************************************/
  /*                                                                                       */
  /* signal declarations                                                                   */
  /*                                                                                       */
  /*****************************************************************************************/
  wire        assem_nxt;                                   /* assem_reg input              */
  wire        b_nxt;                                       /* b register input             */
  wire        cry_one;                                     /* carry, initially set         */
  wire        cry_zer;                                     /* carry, initially clear       */
  wire        data_drv;                                    /* mem/periph data direction    */
  wire        data_out;                                    /* mem/periph data output       */
  wire        fo_bus;                                      /* flag output bus              */
  wire        g_nxt;                                       /* g register input             */
  wire        isa_nxt;                                     /* next isa output data         */
  wire        jmp_cry;                                     /* jmp and carry                */
  wire        jmp_nocry;                                   /* jmp and no carry             */
  wire        ld_cycle;                                    /* load machine cycle           */
  wire        ld_inst;                                     /* load instruction             */
  wire        m_nxt;                                       /* m register input             */
  wire        n_nxt;                                       /* n register input             */
  wire        nosync;                                      /* inhibit sync output          */
  wire        pc_inc;                                      /* pc incrementer               */
  wire        pc_rel;                                      /* pc adder                     */
  wire        ptr_eq;                                      /* pointer p = pointer q        */
  wire        ptr_eqn;                                     /* pointer = test value         */
  wire        state_b;                                     /* decoded states               */
  wire        sync_reg;                                    /* sync signal                  */
  wire        te_a14;                                      /* time enable: addr 14         */
  wire        te_cgo;                                      /* time enable: nibble 1        */
  wire        te_fi;                                       /* time enable: flag input      */
  wire        te_flg;                                      /* time enable: flag            */
  wire        te_fo;                                       /* time enable: flag output     */
  wire        te_g;                                        /* time enable: ptr, ptr+       */
  wire        te_go;                                       /* time enable: nibbles 2,3,4,5 */
  wire        te_ky;                                       /* time enable: nibbles 3,4     */
  wire        te_m;                                        /* time enable: mantissa        */
  wire        te_ms;                                       /* time enable: mant sign       */
  wire        te_p;                                        /* time enable: on ptr          */
  wire        te_pq;                                       /* time enable: p thru q        */
  wire        te_rcrd;                                     /* time enable: rcrd            */
  wire        te_st;                                       /* time enable: nibbles 0,1     */
  wire        te_syn;                                      /* time enable: nibbles 11,12   */
  wire        te_t0;                                       /* time enable: time t0         */
  wire        te_t55;                                      /* time enable: time t55        */
  wire        te_wpt;                                      /* time enable: word thru ptr   */
  wire        te_x;                                        /* time enable: exponent        */
  wire        te_xs;                                       /* time enable: exp sign        */
  wire  [3:0] ptr_reg;                                     /* active pointer               */

  reg         a_sel;                                       /* a register input select      */
  reg         add_inst;                                    /* add instruction              */
  reg         addly_reg;                                   /* address delay                */
  reg         alu_out;                                     /* alu output                   */
  reg         alu_cry;                                     /* alu carry output             */
  reg         b_sel;                                       /* b register input select      */
  reg         c_sel;                                       /* c register input select      */
  reg         chkkb_inst;                                  /* chk kb instruction           */
  reg         cperin_inst;                                 /* peripheral input inst        */
  reg         cperin_reg;                                  /* cperin latched               */
  reg         cregn_inst;                                  /* c=regn instruction           */
  reg         cry_reg;                                     /* carry register               */
  reg         cry_sel;                                     /* carry register input select  */
  reg         cxisa_inst;                                  /* cxisa instruction            */
  reg         cyinc_reg;                                   /* pc incrementer carry         */
  reg         cyjmp_reg;                                   /* pc adder carry               */
  reg         dadj_cry;                                    /* decimal adjusted carry       */
  reg         en_areg;                                     /* a register enable            */
  reg         en_breg;                                     /* b register enable            */
  reg         en_cry;                                      /* carry flag enable            */
  reg         en_greg;                                     /* g register enable            */
  reg         en_hexmd;                                    /* hexmd register enable        */
  reg         en_mreg;                                     /* m register enable            */
  reg         en_nreg;                                     /* n register enable            */
  reg         en_stkreg;                                   /* stk register enable          */
  reg         en_streg;                                    /* st register enable           */
  reg         fo_nxt;                                      /* fo register input            */
  reg         hexmd_reg;                                   /* hex mode                     */
  reg         inicry_reg;                                  /* carry in (inc/dec) register  */
  reg         isa_drv;                                     /* isa drive enable             */
  reg         isa_out;                                     /* isa output data              */
  reg         jmpldi_inst;                                 /* jmp/jsb/ldi/cxisa inst       */
  reg         jmpldi_reg;                                  /* jmp/jsb/ldi/cxisa latched    */
  reg         pc_nxt;                                      /* next pc input                */
  reg         pflagn_inst;                                 /* peripheral flag inst         */
  reg         pflagn_reg;                                  /* pflagn latched               */
  reg         powoff_inst;                                 /* power off instruction        */
  reg         psel_inst;                                   /* ptr select instruction       */
  reg         qsel_reg;                                    /* q/p pointer select           */
  reg         rcrd_inst;                                   /* rcrd instruction             */
  reg         rcrd_reg;                                    /* rcrd time enable             */
  reg         rstkb_inst;                                  /* rst kb instruction           */
  reg         selprf_inst;                                 /* selprf inst                  */
  reg         selprf_reg;                                  /* selprf latched               */
  reg         st_nxt;                                      /* st register input            */
  reg         stk0_nxt, stk1_nxt, stk2_nxt, stk3_nxt;      /* stack reg inputs             */
  reg         subone_ctl;                                  /* frst word of subr decode     */
  reg         subone_reg;                                  /* frst word of subr reg        */
  reg         sync_raw;                                    /* sync signal timing           */
  reg         te_alu;                                      /* time enable: alu             */
  reg         te_cry;                                      /* time enable: carry flag      */
  reg         te_isa;                                      /* time enable: load from isa   */
  reg         te_rr;                                       /* time enable: rr              */
  reg         teg_reg, tpq_reg, twpt_reg;                  /* pointer time selects         */
  reg   [1:0] aludly_reg;                                  /* alu delayed for dadj         */
  reg   [1:0] deepslp_reg;                                 /* deep sleep start             */
  reg   [1:0] inst2_reg;                                   /* second instruction reg       */
  reg   [1:0] ptr_ctl;                                     /* ptr operation decode         */
  reg   [2:0] dadj_out;                                    /* decimal adjusted value       */
  reg   [3:0] a_nxt;                                       /* a register input             */
  reg   [3:0] adly_reg;                                    /* a register delay             */
  reg   [3:0] c_nxt;                                       /* c register input             */
  reg   [3:0] cycle_reg;                                   /* clock cycle                  */
  reg   [3:0] lc_reg;                                      /* lc delay                     */
  reg   [3:0] p_reg;                                       /* pointer p                    */
  reg   [3:0] ptr_nxt;                                     /* pointer next value           */
  reg   [3:0] q_reg;                                       /* pointer q                    */
  reg   [3:0] state_reg;                                   /* machine state                */
  reg   [7:0] assem2_reg;                                  /* second inst assembly reg     */
  reg   [7:0] fo_reg;                                      /* fo register                  */
  reg   [7:0] g_reg;                                       /* g register                   */
  reg   [9:0] assem_reg;                                   /* inst assembly reg            */
  reg   [9:0] inst_reg;                                    /* instruction register         */
  reg  [13:0] st_reg;                                      /* st register                  */
  reg  [15:0] pc_reg;                                      /* program counter              */
  reg  [15:0] stk3_reg, stk2_reg, stk1_reg, stk0_reg;      /* stack registers              */
  reg  [55:0] a_reg;                                       /* a register                   */
  reg  [55:0] b_reg;                                       /* b register                   */
  reg  [55:0] c_reg;                                       /* c register                   */
  reg  [55:0] dout_reg;                                    /* stable data register         */
  reg  [55:0] m_reg;                                       /* m register                   */
  reg  [55:0] n_reg;                                       /* n register                   */

  /*****************************************************************************************/
  /*                                                                                       */
  /* newt parallel interface                                                               */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (posedge ph2) begin
    if (te_t55) dout_reg <= (te_rcrd) ? {c_nxt, c_reg[52:1]} : c_reg;
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* processor state machine                                                               */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (posedge ph2 or negedge rstb) begin
    if                 (!rstb) deepslp_reg <= 2'b00;
    else if (ld_inst || !dpwo) deepslp_reg <= (ld_inst) ? {deepslp_reg[0], 1'b0} : 2'b11;
    end

  always @ (posedge ph2 or negedge rstb) begin
    if (!rstb) cycle_reg <= 4'b0001;
    else       cycle_reg <= {( !cycle_reg[3]   && cycle_reg[2]),
                             (~|cycle_reg[3:2] && cycle_reg[1]),
                              ~|cycle_reg[3:1], cycle_reg[3]};
    end

  always @ (posedge ph2 or negedge rstb) begin
    if         (!rstb) state_reg <= 4'b1111;
    else if (ld_cycle) state_reg <= {(state_reg[2] || state_b), state_reg[1:0],
                                     ((state_reg[0] ^ !state_reg[3]) && !state_b)};
    end

  assign ld_cycle = cycle_reg[3];
  assign ld_inst  = te_t55 && !nosync;

  /*****************************************************************************************/
  /*                                                                                       */
  /* time enables                                                                          */
  /*                                                                                       */
  /*****************************************************************************************/
  assign state_b =  state_reg[3] && &state_reg[1:0];
  assign te_a14  = ~|(state_reg ^ 4'b1010) && cycle_reg[0];
  assign te_cgo  = &state_reg[3:2] && ~|state_reg[1:0];
  assign te_fi   = ~|(state_reg ^ inst_reg[9:6]) && cycle_reg[1];
  assign te_flg  = ~|(state_reg ^ inst_reg[9:6]);
  assign te_fo   = (~|state_reg[3:2] && !state_reg[0]) || (state_reg[3] && !state_reg[0]) ||
                   (!state_reg[3] && !state_reg[1] && state_reg[0]);
  assign te_g    = teg_reg || ~|(ptr_reg ^ state_reg);
  assign te_go   = !state_reg[3] && ((!state_reg[2] && !state_reg[0]) ||
                   (!state_reg[1] && state_reg[0]));
  assign te_ky   = ~|state_reg[3:1];
  assign te_p    = ~|(ptr_reg ^ state_reg);
  assign te_wpt  = twpt_reg;
  assign te_pq   = ~|(p_reg ^ state_reg) || tpq_reg;
  assign te_st   = &state_reg[3:2] && !state_reg[0];
  assign te_syn  = (~|(state_reg ^ 4'b0110) || ~|(state_reg ^ 4'b1101)) && nosync;
  assign te_t0   = &state_reg[3:1] && !state_reg[0] && cycle_reg[0];
  assign te_t55  = state_b && cycle_reg[3];
  assign te_x    = state_reg[3] && (state_reg[2] || !state_reg[1]) && !state_reg[0];
  assign te_xs   = state_reg[3] && ~|state_reg[2:0];
  assign te_m    = !state_reg[3] || (!state_reg[1] && state_reg[0]) ||
                   (!state_reg[2] && state_reg[1] && !state_reg[0]);
  assign te_ms   = state_reg[3] && &state_reg[1:0];

  always @ (inst_reg or te_g or te_go or te_ky or te_m or te_ms or te_p or te_pq or te_st or
            te_wpt or te_x or te_xs) begin
    casex (inst_reg) //synopsys parallel_case
      10'b00x1011000,
      10'b001x011000: te_alu = te_g;                       /* ptr & ptr+                   */
      10'b0101110000,
      10'b0110110000: te_alu = te_go;                      /* stack                        */
      10'b1111001100,
      10'b1000100000: te_alu = te_ky;                      /* keyboard                     */
      10'bxxxxx11010: te_alu = te_m;                       /* mantissa                     */
      10'bxxxxx11110: te_alu = te_ms;                      /* mantissa sign                */
      10'bxxxx010000,
      10'bxxxxx00010: te_alu = te_p;                       /* on pointer                   */
      10'bxxxxx10010: te_alu = te_pq;                      /* p through q                  */
      10'b11x1011000,
      10'b111x011000: te_alu = te_st;                      /* status                       */
      10'bxxxxx01010: te_alu = te_wpt;                     /* word through pointer         */
      10'bx100110000,
      10'bxxxxx00110: te_alu = te_x;                       /* exponent                     */
      10'bxxxxx10110: te_alu = te_xs;                      /* exponent sign                */
      default:        te_alu = 1'b1;                       /* always on (default)          */
      endcase
    end

  always @ (inst_reg or te_fi or te_flg or te_ky or te_m or te_ms or te_p or te_pq or
            te_t0 or te_wpt or te_x or te_xs) begin
    casex (inst_reg) //synopsys parallel_case
      10'bx0xx101100,
      10'bxx0x101100,
      10'bxxx0101100: te_cry = te_fi;                      /* fi                           */
      10'bx0xx001100,
      10'bxx0x001100,
      10'bxxx0001100: te_cry = te_flg;                     /* st                           */
      10'b1111001100: te_cry = te_ky;                      /* keyboard                     */
      10'bxxxxx11010: te_cry = te_m;                       /* mantissa                     */
      10'bxxxxx11110: te_cry = te_ms;                      /* mantissa sign                */
      10'bxxxxx00010: te_cry = te_p;                       /* on pointer                   */
      10'bxxxxx10010: te_cry = te_pq;                      /* p through q                  */
      10'bxxxx100100: te_cry = te_t0;                      /* t0                           */
      10'bxxxxx01010: te_cry = te_wpt;                     /* word through pointer         */
      10'bxxxxx00110: te_cry = te_x;                       /* exponent                     */
      10'bxxxxx10110: te_cry = te_xs;                      /* exponent sign                */
      10'bxxxxx01110: te_cry = 1'b1;                       /* all                          */
      default:        te_cry = 1'b0;                       /* always off (default)         */
      endcase
    end

  always @ (inst_reg or ptr_reg or q_reg or state_reg) begin
    casex (inst_reg[4:2])
      3'bx01:  te_rr = ~|(state_reg ^ 4'b1000);            /* exp or exp sign              */
      3'b0x0:  te_rr = ~|(state_reg ^ ptr_reg);            /* ptr or word thru ptr         */
      3'bx11:  te_rr = ~|(state_reg ^ 4'b1011);            /* word or mant sign            */
      3'b110:  te_rr = ~|(state_reg ^ 4'b1101);            /* mantissa                     */
      default: te_rr = ~|(state_reg ^ q_reg) || ~|(state_reg ^ 4'b1011);
      endcase
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* rcrd time enable                                                                      */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    case (inst_reg[5:0])
      6'b111100: rcrd_inst = 1'b1;
      default:   rcrd_inst = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    if (ld_cycle) rcrd_reg <= ld_inst || (rcrd_reg && !(rcrd_inst && te_flg));
    end

  assign te_rcrd = rcrd_reg && !(rcrd_inst && te_flg);

  /*****************************************************************************************/
  /*                                                                                       */
  /* hex/decimal mode                                                                      */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    case (inst_reg)
      10'b1010100000,
      10'b1001100000: en_hexmd = 1'b1;
      default:        en_hexmd = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    if (en_hexmd) hexmd_reg <= inst_reg[6];
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* keyboard interface                                                                    */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (assem_reg) begin
    case (assem_reg)
      10'b1111001100: chkkb_inst = 1'b1;
      default:        chkkb_inst = 1'b0;
      endcase
    end

  always @ (assem_reg) begin
    case (assem_reg)
      10'b1111001000: rstkb_inst = 1'b1;
      default:        rstkb_inst = 1'b0;
      endcase
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* power off decode                                                                      */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    case (inst_reg)
      10'b0001100000: powoff_inst = 1'b1;
      default:        powoff_inst = 1'b0;
      endcase
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* first subroutine byte identifier (for auto-return)                                    */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or inst2_reg or jmp_cry or jmp_nocry) begin
    case ({inst_reg[1:0], inst2_reg})
      4'b0100: subone_ctl = jmp_nocry;                     /* push pc+ if no carry         */
      4'b0101: subone_ctl = jmp_cry;                       /* push pc+ if carry            */
      default: subone_ctl = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    if (ld_inst) subone_reg <= subone_ctl;
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* instruction register                                                                  */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or te_go or te_x) begin
    casex (inst_reg)
      10'bx100110000: te_isa = te_x;
      10'bxxxxxxxxx1: te_isa = te_go;
      default:        te_isa = 1'b0;
      endcase
    end

  assign assem_nxt = (sync_raw) ? isa_in : (assem_reg[9] && inst_reg[0]);

  always @ (posedge ph2 or negedge rstb) begin
    if                                (!rstb) assem_reg <= 10'h000;
    else if (sync_raw || (te_isa && !nosync)) assem_reg <= {assem_nxt, assem_reg[9:1]};
    end

  always @ (posedge ph2 or negedge rstb) begin
    if                              (!rstb) assem2_reg <= 8'h00;
    else if (te_syn || (te_isa && !nosync)) assem2_reg <= {assem_reg[2], assem2_reg[7:1]};
    end

  always @ (posedge ph2 or negedge rstb) begin
    if        (!rstb) inst_reg <= 10'h000;
    else if (ld_inst) inst_reg <= assem_reg;
    end

  always @ (posedge ph2 or negedge rstb) begin
    if                 (!rstb) inst2_reg <= 2'b00;
    else if (te_t55 && nosync) inst2_reg <= assem_reg[1:0];
    end

  always @ (posedge ph2) begin
    if (ld_inst || te_p) lc_reg <= (ld_inst) ? assem_reg[9:6] : {1'b0, lc_reg[3:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* special peripheral instructions                                                       */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (assem_reg) begin
    case (assem_reg[5:0])
      6'b100100: selprf_inst = 1'b1;                      /* selprf instruction            */
      default:   selprf_inst = 1'b0;
      endcase
    end

  always @ (assem_reg) begin
    case (assem_reg[5:1])
      5'b11101: cperin_inst = 1'b1;                       /* c=data(np) instruction        */
      default:  cperin_inst = 1'b0;
      endcase
    end

  always @ (assem_reg) begin
    case (assem_reg[5:1])
      5'b00001: pflagn_inst = 1'b1;                       /* if pflagn=1 instruction       */
      default:  pflagn_inst = 1'b0;
      endcase
    end

  always @ (posedge ph2 or negedge rstb) begin
    if (!rstb) selprf_reg <= 1'b0;
    else       selprf_reg <= ((ld_inst && selprf_inst) || selprf_reg) &&
                             !(te_t55 && assem_reg[0]);
    end

  always @ (posedge ph2) begin
    if (te_t55) begin
      cperin_reg <= selprf_reg && cperin_inst;
      pflagn_reg <= selprf_reg && pflagn_inst;
      end
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* program counter                                                                       */
  /*                                                                                       */
  /*****************************************************************************************/
  assign jmp_cry   = !nosync &&  cry_reg;
  assign jmp_nocry = !nosync && !cry_reg;
  assign pc_inc    = pc_reg[0] ^ cyinc_reg;
  assign pc_rel    = pc_reg[0] ^ cyjmp_reg ^ assem_reg[3];

  always @ (inst_reg or inst2_reg or assem2_reg or c_reg or cry_reg or jmp_cry or
            jmp_nocry or ky_data or nosync or pc_inc or pc_rel or pc_reg or stk0_reg or
            subone_reg or te_go or te_ky) begin
    casex ({inst_reg, inst2_reg})
      12'b0000000000xx: pc_nxt = (subone_reg) ? stk0_reg      : pc_inc;
      12'b0111100000xx: pc_nxt = (te_go)      ? c_reg[0]      : pc_reg[0];
      12'b1000110000xx: pc_nxt = (te_ky)      ? ky_data       : pc_reg[0];
      12'b1100110000xx: pc_nxt = (nosync)     ? pc_reg[0]     : pc_inc;
      12'b1101100000xx: pc_nxt = (cry_reg)    ? stk0_reg[0]   : pc_inc;
      12'b1110100000xx: pc_nxt = (cry_reg)    ? pc_inc        : stk0_reg[0];
      12'b1111100000xx: pc_nxt = stk0_reg[0];
      12'bxxxxxxx011xx: pc_nxt = (cry_reg)    ? pc_inc        : pc_rel;
      12'bxxxxxxx111xx: pc_nxt = (cry_reg)    ? pc_rel        : pc_inc;
      12'bxxxxxxxx01x0: pc_nxt = (jmp_nocry)  ? assem2_reg[0] : pc_inc;
      12'bxxxxxxxx01x1: pc_nxt = (jmp_cry)    ? assem2_reg[0] : pc_inc;
      default:          pc_nxt = pc_inc;
      endcase
    end

  always @ (posedge ph2) begin
    if (te_cgo || te_go) begin
      cyinc_reg <= te_cgo || (cyinc_reg && pc_reg[0]);
      cyjmp_reg <= !te_cgo && ((assem_reg[3] && cyjmp_reg) ||
                               (pc_reg[0] && (assem_reg[3] || cyjmp_reg)));
      end
    end

  always @ (posedge ph2 or negedge rstb) begin
    if      (!rstb) pc_reg <= 16'hffff;
    else if (te_go) pc_reg <= {(pc_nxt || powoff_inst), pc_reg[15:1]};
    end

  always @ (posedge ph2) begin
    addly_reg <= (cxisa_inst && nosync) ? c_reg[0] : pc_nxt;
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* isa output                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  assign isa_nxt = addly_reg;

  always @ (inst_reg) begin
    case (inst_reg)
      10'b1100110000: cxisa_inst = 1'b1;                   /* cxisa instruction            */
      default:        cxisa_inst = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    isa_out <= addly_reg;
    end

  always @ (posedge ph2 or negedge rstb) begin
    if (!rstb) isa_drv <= 1'b0;
    else       isa_drv <= (~|state_reg && cycle_reg[1]) ||
                          (!(~|(state_reg ^ 4'b1010) && cycle_reg[1]) && isa_drv);
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* data bus                                                                              */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    case (inst_reg[5:0])
      6'b111000: cregn_inst = 1'b1;                       /* c=regn instruction           */
      default:   cregn_inst = 1'b0;
      endcase
    end

  assign data_drv = !cregn_inst && !cperin_reg;
  assign data_out = c_reg[0];

  /*****************************************************************************************/
  /*                                                                                       */
  /* sync output                                                                           */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (assem_reg) begin
    casex (assem_reg)
      10'b0100110000,                                     /* ldi instruction               */
      10'b1100110000,                                     /* cxisa instruction             */
      10'bxxxxxxxx01: jmpldi_inst = 1'b1;                 /* jmp or jsb instruction        */
      default:        jmpldi_inst = 1'b0;
      endcase
    end

  always @ (posedge ph2 or negedge rstb) begin
    if       (!rstb) jmpldi_reg <= 1'b0;
    else if (te_t55) jmpldi_reg <= jmpldi_inst && !jmpldi_reg && !selprf_reg;
    end

  assign nosync   = jmpldi_reg || selprf_reg;
  assign sync_reg = sync_raw && !nosync;

  always @ (posedge ph2 or negedge rstb) begin
    if (!rstb) sync_raw <= 1'b0;
    else       sync_raw <= (~|(state_reg ^ 4'b0011) && cycle_reg[3]) ||
                           (!(state_b && cycle_reg[1]) && sync_raw);
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* stack                                                                                 */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or inst2_reg or cry_reg or nosync or subone_reg) begin
    casex ({inst_reg, inst2_reg})
      12'b0000000000xx: en_stkreg = subone_reg;
      12'b1101100000xx: en_stkreg =  cry_reg;
      12'b1110100000xx: en_stkreg = !cry_reg;
      12'bxxxxxxxx0100: en_stkreg = !cry_reg && !nosync;
      12'bxxxxxxxx0101: en_stkreg =  cry_reg && !nosync;
      12'b0000100000xx,
      12'b1111100000xx,
      12'b0101110000xx,
      12'b0110110000xx: en_stkreg = 1'b1;
      default:          en_stkreg = 1'b0;
      endcase
    end

  always @ (inst_reg or stk2_reg) begin
    casex (inst_reg)
      10'bxxxxxxxx01,
      10'b0101110000: stk3_nxt = stk2_reg[0];
      default:        stk3_nxt = 1'b0;
      endcase
    end

  always @ (inst_reg or stk1_reg or stk3_reg) begin
    casex (inst_reg)
      10'bxxxxxxxx01,
      10'b0101110000: stk2_nxt = stk1_reg[0];
      default:        stk2_nxt = stk3_reg[0];
      endcase
    end

  always @ (inst_reg or stk0_reg or stk2_reg) begin
    casex (inst_reg)
      10'bxxxxxxxx01,
      10'b0101110000: stk1_nxt = stk0_reg[0];
      default:        stk1_nxt = stk2_reg[0];
      endcase
    end

  always @ (inst_reg or c_reg or pc_inc or stk1_reg) begin
    casex (inst_reg)
      10'bxxxxxxxx01: stk0_nxt = pc_inc;
      10'b0101110000: stk0_nxt = c_reg[0];
      default:        stk0_nxt = stk1_reg[0];
      endcase
    end

  always @ (posedge ph2) begin
    if (te_go && en_stkreg) begin
      stk3_reg <= {stk3_nxt, stk3_reg[15:1]};
      stk2_reg <= {stk2_nxt, stk2_reg[15:1]};
      stk1_reg <= {stk1_nxt, stk1_reg[15:1]};
      stk0_reg <= {stk0_nxt, stk0_reg[15:1]};
      end
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* pointer select                                                                        */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b001x100000: psel_inst = 1'b1;                    /* select pointer               */
      default:        psel_inst = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    if (ld_inst && psel_inst) qsel_reg <= inst_reg[6];
    end

  assign ptr_reg = (qsel_reg) ? q_reg : p_reg;

  /*****************************************************************************************/
  /*                                                                                       */
  /* pointers                                                                              */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg) //synopsys parallel_case
      10'bxxx0011100,
      10'bxx0x011100,
      10'bx0xx011100: ptr_ctl = 2'b01;                     /* load to pointer              */
      10'bxxxx010000,
      10'b1111010100: ptr_ctl = 2'b10;                     /* decrement pointer            */
      10'b1111011100: ptr_ctl = 2'b11;                     /* increment pointer            */
      default:        ptr_ctl = 2'b00;
      endcase
    end

  assign ptr_eq  = ~|(p_reg ^ q_reg);
  assign ptr_eqn = ~|(ptr_reg ^ inst_reg[9:6]);

  always @ (ptr_ctl or inst_reg or ptr_reg) begin
    case (ptr_ctl)
      2'b10:   ptr_nxt = {((ptr_reg[1] ^ !ptr_reg[0]) || &ptr_reg[3:1]),
                          (ptr_reg[3] && ~&ptr_reg[2:1]), ptr_reg[2:1]};
      2'b11:   ptr_nxt = {(ptr_reg[2] || (ptr_reg[3] && &ptr_reg[1:0])), ptr_reg[1:0],
                          ((!ptr_reg[3] && !ptr_reg[0]) ||
                          (ptr_reg[3] && !ptr_reg[1] && ptr_reg[0]))};
      default: ptr_nxt = inst_reg[9:6];
      endcase
    end

  always @ (posedge ph2) begin
    if (ld_cycle) begin
      teg_reg  <= ~|(ptr_reg ^ state_reg);
      tpq_reg  <= |(q_reg ^ state_reg) && !ld_inst && (~|(p_reg ^ state_reg) || tpq_reg);
      twpt_reg <= ld_inst || (|(ptr_reg ^ state_reg) && twpt_reg);
      end
    end

  always @ (posedge ph2) begin
    if (ld_inst && |ptr_ctl && !qsel_reg) p_reg <= ptr_nxt;
    if (ld_inst && |ptr_ctl &&  qsel_reg) q_reg <= ptr_nxt;
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* a register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b0110100000,
      10'b11xx110000,
      10'bxxxxxxxx10: en_areg = 1'b1;
      default:        en_areg = 1'b0;
      endcase
    end

  always @ (inst_reg or a_reg or adly_reg or alu_out or b_reg or c_reg or te_rr) begin
    casex (inst_reg)
      10'b0110100000,
      10'b00000xxx10: a_sel = 1'b0;
      10'b00011xxx10: a_sel = b_reg[0];
      10'b00101xxx10,
      10'b01000xxx10: a_sel = c_reg[0];
      10'b01001xxx10,
      10'b0101xxxx10,
      10'b0110xxxx10,
      10'b01110xxx10: a_sel = alu_out;
      10'b11100xxx10: a_sel = !te_rr && a_reg[4];
      10'b11111xxx10: a_sel = adly_reg[0];
      default:        a_sel = a_reg[0];
      endcase
    end

  always @ (inst_reg or a_reg or a_sel or alu_out or dadj_out or ld_cycle or hexmd_reg or
            te_alu) begin
    casex (inst_reg)
      10'b01001xxx10,
      10'b0101xxxx10,
      10'b0110xxxx10,
      10'b01110xxx10: begin
        casex ({te_alu, ld_cycle, hexmd_reg})
          3'b10x,
          3'b111:     a_nxt = {alu_out,  a_reg[55:53]};
          3'b110:     a_nxt = {dadj_out, a_reg[53]};
          default:    a_nxt = {a_reg[0], a_reg[55:53]};
          endcase
        end
      default:        a_nxt = (te_alu) ? {a_sel, a_reg[55:53]} : {a_reg[0], a_reg[55:53]};
      endcase
    end

  always @ (posedge ph2) begin
    if (en_areg) a_reg <= {a_nxt, a_reg[52:1]};
    if (ld_inst || te_alu) adly_reg <= (ld_inst) ? 4'b0000 : {a_reg[0], adly_reg[3:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* b register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b0110100000,
      10'bxxxxxxxx10: en_breg = 1'b1;
      default:        en_breg = 1'b0;
      endcase
    end

  always @ (inst_reg or a_reg or b_reg or c_reg or te_rr) begin
    casex (inst_reg)
      10'b0110100000,
      10'b00001xxx10: b_sel = 1'b0;
      10'b00011xxx10,
      10'b00100xxx10: b_sel = a_reg[0];
      10'b00111xxx10: b_sel = c_reg[0];
      10'b11101xxx10: b_sel = !te_rr && b_reg[4];
      default:        b_sel = b_reg[0];
      endcase
    end

  assign b_nxt = (te_alu) ? b_sel : b_reg[0];

  always @ (posedge ph2) begin
    if (en_breg) b_reg <= {b_nxt, b_reg[55:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* c register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or a_reg or alu_out or assem_reg or b_reg or c_reg or cperin_reg or
            data_in or g_reg or ky_data or lc_reg or m_reg or n_reg or st_reg or stk0_reg or
            te_rr) begin
    casex (inst_reg)
      10'b1000100000: c_sel = ky_data;
      10'bxxxx010000: c_sel = lc_reg[0];
      10'b001x011000: c_sel = g_reg[0];
      10'b011x011000: c_sel = m_reg[0];
      10'b111x011000: c_sel = st_reg[0];
      10'b0110100000,
      10'b00010xxx10: c_sel = 1'b0;
      10'b001x110000: c_sel = n_reg[0];
      10'b0110110000: c_sel = stk0_reg[0];
      10'bx100110000: c_sel = assem_reg[0];
      10'bxxxx111000: c_sel = data_in;
      10'bxxxx100100: c_sel = (cperin_reg) ? data_in : c_reg[0];
      10'b00101xxx10: c_sel = a_reg[0];
      10'b0011xxxx10: c_sel = b_reg[0];
      10'b1101110000,
      10'b1110110000,
      10'b01111xxx10,
      10'b100xxxxx10,
      10'b1010xxxx10: c_sel = alu_out;
      10'b11110xxx10: c_sel = !te_rr && c_reg[4];
      default:        c_sel = c_reg[0];
      endcase
    end

  always @ (inst_reg or alu_out or c_reg or c_sel or dadj_out or ld_cycle or hexmd_reg or
            te_alu) begin
    casex (inst_reg)
      10'b100xxxxx10,
      10'b1010xxxx10,
      10'b01111xxx10: begin
        casex ({te_alu, ld_cycle, hexmd_reg})
          3'b10x,
          3'b111:     c_nxt = {alu_out,  c_reg[55:53]};
          3'b110:     c_nxt = {dadj_out, c_reg[53]};
          default:    c_nxt = {c_reg[0], c_reg[55:53]};
          endcase
        end
      default:        c_nxt = (te_alu) ? {c_sel, c_reg[55:53]} : {c_reg[0], c_reg[55:53]};
      endcase
    end

  always @ (posedge ph2) begin
    if (te_rcrd) c_reg <= {c_nxt, c_reg[52:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* fo register                                                                           */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or fo_reg or st_reg) begin
    casex (inst_reg)
      10'b10x1011000: fo_nxt = st_reg[0];
      default:        fo_nxt = fo_reg[0];
      endcase
    end

  always @ (posedge ph2 or negedge rstb) begin
    if                  (!rstb) fo_reg <= 8'h00;
    else if (te_fo && ld_cycle) fo_reg <= {fo_nxt, fo_reg[7:1]};
    end

  assign fo_bus = fo_reg[0];

  /*****************************************************************************************/
  /*                                                                                       */
  /* g register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b00xx011000: en_greg = 1'b1;
      default:        en_greg = 1'b0;
      endcase
    end

  assign g_nxt = (inst_reg[6]) ? c_reg[0] : g_reg[0];

  always @ (posedge ph2) begin
    if (en_greg && te_g) g_reg <= {g_nxt, g_reg[7:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* m register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b01xx011000: en_mreg = 1'b1;
      default:        en_mreg = 1'b0;
      endcase
    end

  assign m_nxt = (inst_reg[6]) ? c_reg[0] : m_reg[0];

  always @ (posedge ph2) begin
    if (en_mreg) m_reg <= {m_nxt, m_reg[55:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* n register                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    casex (inst_reg)
      10'b00xx110000: en_nreg = 1'b1;
      default:        en_nreg = 1'b0;
      endcase
    end

  assign n_nxt = (inst_reg[6]) ? c_reg[0] : n_reg[0];

  always @ (posedge ph2) begin
    if (en_nreg) n_reg <= {n_nxt, n_reg[55:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* st register                                                                           */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or ld_cycle or te_st or te_fo) begin
    casex (inst_reg)
      10'bxxxx000100,
      10'bxxxx001000,
      10'bxxxx001100,
      10'b10xx011000: en_streg = ld_cycle;
      10'b11xx011000: en_streg = te_st || (!te_fo && ld_cycle);
      default:        en_streg = 1'b0;
      endcase
    end

  always @ (inst_reg or c_reg or fo_reg or st_reg or te_fo or te_flg or te_st) begin
    casex (inst_reg) //synopsys parallel_case
      10'bxxx0000100,
      10'bxx0x000100,
      10'bx0xx000100,
      10'b0xxx000100: st_nxt = st_reg[0] && !te_flg;
      10'b1111000100: st_nxt = st_reg[0] && !te_fo;
      10'bxxxx001000: st_nxt = st_reg[0] ||  te_flg;
      10'b101x011000: st_nxt = (te_fo) ? fo_reg[0] : st_reg[0];
      10'b11x1011000: st_nxt = (te_st) ? c_reg[0]  : st_reg[0];
      default:        st_nxt = st_reg[0];
      endcase
    end

  always @ (posedge ph2) begin
    if (en_streg || deepslp_reg[1]) st_reg <= (deepslp_reg[1]) ? 14'h0000 :
                                                                 {st_nxt, st_reg[13:1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* alu                                                                                   */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or a_reg or b_reg or c_reg or cry_one or cry_zer) begin
    casex (inst_reg)
      10'b1101110000: alu_out = a_reg[0] || c_reg[0];
      10'b1110110000: alu_out = a_reg[0] && c_reg[0];
      10'b01001xxx10,
      10'b01100xxx10,
      10'b11001xxx10: alu_out = a_reg[0] ^ b_reg[0] ^ cry_zer;
      10'b01x10xxx10,
      10'b100x0xxx10,
      10'b11000xxx10: alu_out = a_reg[0] ^ c_reg[0] ^ cry_zer;
      10'b01011xxx10,
      10'b01101xxx10: alu_out = a_reg[0] ^ cry_one;
      10'b100x1xxx10,
      10'b10101xxx10: alu_out = c_reg[0] ^ cry_one;
      10'b10100xxx10: alu_out = c_reg[0] ^ cry_zer;
      10'b01111xxx10: alu_out = cry_zer;
      default:        alu_out = 1'b0;
      endcase
    end

  always @ (inst_reg or a_reg or b_reg or c_reg or cry_one or cry_zer) begin
    casex (inst_reg)
      10'b01001xxx10: alu_cry = (b_reg[0] && cry_zer) ||
                                ( a_reg[0] && (b_reg[0] || cry_zer));
      10'b01010xxx10,
      10'b10000xxx10: alu_cry = (c_reg[0] && cry_zer) ||
                                ( a_reg[0] && (c_reg[0] || cry_zer));
      10'b01011xxx10: alu_cry =  a_reg[0] && cry_one;
      10'b01100xxx10,
      10'b11001xxx10: alu_cry = (b_reg[0] && cry_zer) ||
                                (!a_reg[0] && (b_reg[0] || cry_zer));
      10'b01101xxx10: alu_cry = !a_reg[0] && cry_one;
      10'b01110xxx10,
      10'b10010xxx10,
      10'b11000xxx10: alu_cry = (c_reg[0] && cry_zer) ||
                                (!a_reg[0] && (c_reg[0] || cry_zer));
      10'b01111xxx10: alu_cry =  c_reg[0];
      10'b10001xxx10: alu_cry =  c_reg[0] && cry_one;
      10'b10011xxx10: alu_cry = !c_reg[0] && cry_one;
      10'b10111xxx10: alu_cry =  c_reg[0] || cry_zer;
      10'b10101xxx10: alu_cry =  c_reg[0] || cry_one;
      10'b10100xxx10: alu_cry =  c_reg[0] || cry_zer;
      10'b10110xxx10: alu_cry =  b_reg[0] || cry_zer;
      10'b11010xxx10: alu_cry =  a_reg[0] || cry_zer;
      10'b11011xxx10: alu_cry = (a_reg[0] ^ c_reg[0]) || cry_zer;
      default:        alu_cry = 1'b0;
      endcase
    end

  always @ (posedge ph2) begin
    aludly_reg <= {alu_out, aludly_reg[1]};
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* decimal adjust                                                                        */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg) begin
    case (inst_reg[9:5])
      5'b01001,
      5'b01010,
      5'b01011,
      5'b01111,
      5'b10000,
      5'b10001: add_inst = 1'b1;
      default:  add_inst = 1'b0;
      endcase
    end

  always @ (add_inst or alu_cry or alu_out or aludly_reg) begin
    casex ({add_inst, alu_cry, alu_out, aludly_reg})
      5'b01000: dadj_out = 3'b101;                         /* for subtract operation       */
      5'b01001: dadj_out = 3'b110;                         /* if result < 0000             */
      5'b01010: dadj_out = 3'b111;                         /* then add 1010 to result      */
      5'b01011: dadj_out = 3'b000;
      5'b01100: dadj_out = 3'b001;
      5'b01101: dadj_out = 3'b010;
      5'b01110: dadj_out = 3'b011;
      5'b01111: dadj_out = 3'b100;
      5'b1x101: dadj_out = 3'b000;                         /* for add operation            */
      5'b1x110: dadj_out = 3'b001;                         /* if result >= 1010            */
      5'b1x111: dadj_out = 3'b010;                         /* then sub 1010 from result    */
      5'b11000: dadj_out = 3'b011;
      5'b11001: dadj_out = 3'b100;
      5'b11010: dadj_out = 3'b101;
      5'b11011: dadj_out = 3'b110;
      5'b11100: dadj_out = 3'b111;
      default:  dadj_out = {alu_out, aludly_reg};          /* else no change               */
      endcase
    end

  always @ (add_inst or alu_cry or alu_out or aludly_reg) begin
    casex ({add_inst, alu_cry, alu_out, aludly_reg})
      5'b00xxx,
      5'b100xx,
      5'b10100: dadj_cry = 1'b0;
      default:  dadj_cry = 1'b1;
      endcase
    end

  /*****************************************************************************************/
  /*                                                                                       */
  /* carry flag                                                                            */
  /*                                                                                       */
  /*****************************************************************************************/
  always @ (inst_reg or deepslp_reg or pflagn_reg) begin
    casex (inst_reg) //synopsys parallel_case
      10'b1111001100,
      10'bx0xxx01100,
      10'bxx0xx01100,
      10'bxxx0x01100,
      10'bxxxxxxxx10: en_cry = !deepslp_reg[1];
      10'bxxxx100100: en_cry = !deepslp_reg[1] && pflagn_reg;
      default:        en_cry = 1'b0;
      endcase
    end

  always @ (inst_reg or alu_cry or dadj_cry or fi_bus or hexmd_reg or isa_in or ky_flag or
            ld_cycle or lld or ptr_eq or ptr_eqn or st_reg) begin
    casex (inst_reg) //synopsys parallel_case
      10'bx0xx001100,
      10'bxx0x001100,
      10'bxxx0001100: cry_sel = st_reg[0];
      10'b1111001100: cry_sel = ky_flag;
      10'bxxxx010100: cry_sel = ptr_eqn;
      10'b0100100000: cry_sel = ptr_eq;
      10'bxxxx100100: cry_sel = isa_in;
      10'bx0xx101100,
      10'bxx0x101100,
      10'bxxx0101100: cry_sel = fi_bus;
      10'b0101100000: cry_sel = !lld;
      10'b01001xxx10,
      10'b0101xxxx10,
      10'b011xxxxx10,
      10'b100xxxxx10,
      10'b1010xxxx10,
      10'b1100xxxx10: cry_sel = (ld_cycle && !hexmd_reg) ? dadj_cry : alu_cry;
      10'b1011xxxx10,
      10'b1101xxxx10: cry_sel = alu_cry;
      default:        cry_sel = 1'b0;
      endcase
    end

  always @ (posedge ph2 or negedge rstb) begin
    if (!rstb) inicry_reg <= 1'b0;
    else       inicry_reg <= ld_inst || (!te_cry && inicry_reg);
    end

  always @ (posedge ph2 or negedge rstb) begin
    if                               (!rstb) cry_reg <= 1'b0;
    else if (te_cry || (ld_inst && !en_cry)) cry_reg <= cry_sel || deepslp_reg[1];
    end

  assign cry_one = cry_reg ||  inicry_reg;
  assign cry_zer = cry_reg && !inicry_reg;

  endmodule








